@@ -48,7 +48,7 @@
 #define DEFAULT_STACK_SIZE 65536
 /* #define DEBUG_STACK_SIZE */
 
-/* I/O buffer size: 18432 (0x4800) is the maximum size of SSL record payload */
+/* I/O buffer size: 18432 (0x4800) is the maximum size of TLS record payload */
 #define BUFFSIZE 18432
 
 /* how many bytes of random input to read from files for PRNG */
@@ -462,7 +462,6 @@ extern char *sys_errlist[];
 
 #include <openssl/lhash.h>
 #include <openssl/ssl.h>
-#include <openssl/ssl23.h>
 #include <openssl/ui.h>
 #include <openssl/err.h>
 #include <openssl/crypto.h> /* for CRYPTO_* and SSLeay_version */
