@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -46,7 +46,6 @@
 /* #define DEBUG_UCONTEXT */
 
 NOEXPORT void s_poll_realloc(s_poll_set *);
-NOEXPORT int get_socket_error(const SOCKET);
 
 /**************************************** s_poll functions */
 
@@ -66,7 +65,7 @@ void s_poll_free(s_poll_set *fds) {
 
 void s_poll_init(s_poll_set *fds) {
     fds->nfds=0;
-    fds->allocated=4; /* prealloc 4 file desciptors */
+    fds->allocated=4; /* prealloc 4 file descriptors */
     s_poll_realloc(fds);
 }
 
@@ -148,6 +147,15 @@ int s_poll_rdhup(s_poll_set *fds, SOCKET fd) {
     return 0; /* not listed in fds */
 }
 
+int s_poll_err(s_poll_set *fds, SOCKET fd) {
+    unsigned i;
+
+    for(i=0; i<fds->nfds; i++)
+        if(fds->ufds[i].fd==fd)
+            return fds->ufds[i].revents&POLLERR;
+    return 0; /* not listed in fds */
+}
+
 NOEXPORT void s_poll_realloc(s_poll_set *fds) {
     fds->ufds=str_realloc(fds->ufds, fds->allocated*sizeof(struct pollfd));
 }
@@ -352,7 +360,7 @@ void s_poll_free(s_poll_set *fds) {
 
 void s_poll_init(s_poll_set *fds) {
 #ifdef USE_WIN32
-    fds->allocated=4; /* prealloc 4 file desciptors */
+    fds->allocated=4; /* prealloc 4 file descriptors */
 #endif
     s_poll_realloc(fds);
     FD_ZERO(fds->irfds);
@@ -386,27 +394,35 @@ void s_poll_remove(s_poll_set *fds, SOCKET fd) {
 }
 
 int s_poll_canread(s_poll_set *fds, SOCKET fd) {
-    return FD_ISSET(fd, fds->orfds) || FD_ISSET(fd, fds->oxfds);
+    /* ignore exception if there is no error (WinCE 6.0 anomaly) */
+    return FD_ISSET(fd, fds->orfds) ||
+        (FD_ISSET(fd, fds->oxfds) && get_socket_error(fd));
 }
 
 int s_poll_canwrite(s_poll_set *fds, SOCKET fd) {
-    return FD_ISSET(fd, fds->owfds) || FD_ISSET(fd, fds->oxfds);
+    /* ignore exception if there is no error (WinCE 6.0 anomaly) */
+    return FD_ISSET(fd, fds->owfds) ||
+        (FD_ISSET(fd, fds->oxfds) && get_socket_error(fd));
 }
 
 int s_poll_hup(s_poll_set *fds, SOCKET fd) {
     (void)fds; /* squash the unused parameter warning */
     (void)fd; /* squash the unused parameter warning */
-    return 0; /* FIXME: how to detect HUP condition with select()? */
+    return 0; /* FIXME: how to detect the HUP condition with select()? */
 }
 
 int s_poll_rdhup(s_poll_set *fds, SOCKET fd) {
     (void)fds; /* squash the unused parameter warning */
     (void)fd; /* squash the unused parameter warning */
-    return 0; /* FIXME: how to detect RDHUP condition with select()? */
+    return 0; /* FIXME: how to detect the RDHUP condition with select()? */
+}
+
+int s_poll_err(s_poll_set *fds, SOCKET fd) {
+    return FD_ISSET(fd, fds->oxfds);
 }
 
 #ifdef USE_WIN32
-#define FD_SIZE(fds) (sizeof(u_int)+(fds)->allocated*sizeof(SOCKET))
+#define FD_SIZE(fds) (8+(fds)->allocated*sizeof(SOCKET))
 #else
 #define FD_SIZE(fds) (sizeof(fd_set))
 #endif
@@ -418,7 +434,11 @@ int s_poll_wait(s_poll_set *fds, int sec, int msec) {
     do { /* skip "Interrupted system call" errors */
         memcpy(fds->orfds, fds->irfds, FD_SIZE(fds));
         memcpy(fds->owfds, fds->iwfds, FD_SIZE(fds));
+#ifndef _WIN32_WCE
         memcpy(fds->oxfds, fds->ixfds, FD_SIZE(fds));
+#else /* WinCE reports unexpected permanent exceptions */
+        FD_ZERO(fds->oxfds);
+#endif
         if(sec<0) { /* infinite timeout */
             tv_ptr=NULL;
         } else {
@@ -465,7 +485,7 @@ void s_poll_dump(s_poll_set *fds, int level) {
 
 int set_socket_options(SOCKET s, int type) {
     SOCK_OPT *ptr;
-    extern SOCK_OPT sock_opts[];
+    extern SOCK_OPT *sock_opts;
     static char *type_str[3]={"accept", "local", "remote"};
     socklen_t opt_size;
     int retval=0; /* no error found */
@@ -498,17 +518,15 @@ int set_socket_options(SOCKET s, int type) {
                 retval=-1; /* failed to set this option */
             }
         }
-#ifdef DEBUG_FD_ALLOC
         else {
             s_log(LOG_DEBUG, "Option %s set on %s socket",
                 ptr->opt_str, type_str[type]);
         }
-#endif /* DEBUG_FD_ALLOC */
     }
     return retval; /* returns 0 when all options succeeded */
 }
 
-NOEXPORT int get_socket_error(const SOCKET fd) {
+int get_socket_error(const SOCKET fd) {
     int err;
     socklen_t optlen=sizeof err;
 
