@@ -1,6 +1,6 @@
 /*
- *   stunnel       Universal SSL tunnel
- *   Copyright (C) 1998-2014 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   stunnel       TLS offloading and load-balancing proxy
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -54,24 +54,32 @@ int main(int argc, char* argv[]) { /* execution begins here 8-) */
 #ifdef M_MMAP_THRESHOLD
     mallopt(M_MMAP_THRESHOLD, 4096);
 #endif
-    str_init(); /* initialize per-thread string management */
+    tls_init(); /* initialize thread-local storage */
     retval=main_unix(argc, argv);
     main_cleanup();
     return retval;
 }
 
 NOEXPORT int main_unix(int argc, char* argv[]) {
+    int configure_status;
+
 #if !defined(__vms) && !defined(USE_OS2)
     int fd;
 
     fd=open("/dev/null", O_RDWR); /* open /dev/null before chroot */
-    if(fd<0)
+    if(fd==INVALID_SOCKET)
         fatal("Could not open /dev/null");
 #endif
     main_init();
-    if(main_configure(argc>1 ? argv[1] : NULL, argc>2 ? argv[2] : NULL)) {
+    configure_status=main_configure(argc>1 ? argv[1] : NULL,
+        argc>2 ? argv[2] : NULL);
+    switch(configure_status) {
+    case 1: /* error -> exit with 1 to indicate error */
         close(fd);
         return 1;
+    case 2: /* information printed -> exit with 0 to indicate success */
+        close(fd);
+        return 0;
     }
     if(service_options.next) { /* there are service sections -> daemon mode */
 #if !defined(__vms) && !defined(USE_OS2)
@@ -101,6 +109,7 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
 #endif
         daemon_loop();
     } else { /* inetd mode */
+        CLI *c;
 #if !defined(__vms) && !defined(USE_OS2)
         close(fd);
 #endif /* standard Unix */
@@ -108,7 +117,11 @@ NOEXPORT int main_unix(int argc, char* argv[]) {
         signal(SIGCHLD, SIG_IGN); /* ignore dead children */
         signal(SIGPIPE, SIG_IGN); /* ignore broken pipe */
 #endif
-        client_main(alloc_client_session(&service_options, 0, 1));
+        set_nonblock(0, 1); /* stdin */
+        set_nonblock(1, 1); /* stdout */
+        c=alloc_client_session(&service_options, 0, 1);
+        tls_alloc(c, ui_tls, NULL);
+        client_main(c);
     }
     return 0;
 }
@@ -151,6 +164,7 @@ NOEXPORT int daemonize(int fd) { /* go to background */
         exit(0);
     }
 #endif
+    tls_alloc(NULL, ui_tls, "main"); /* reuse thread-local storage */
 #ifdef HAVE_SETSID
     setsid(); /* ignore the error */
 #endif
@@ -212,12 +226,12 @@ void ui_config_reloaded(void) {
 #ifdef ICON_IMAGE
 
 ICON_IMAGE load_icon_default(ICON_TYPE icon) {
-    (void)icon; /* skip warning about unused parameter */
+    (void)icon; /* squash the unused parameter warning */
     return (ICON_IMAGE)0;
 }
 
 ICON_IMAGE load_icon_file(const char *file) {
-    (void)file; /* skip warning about unused parameter */
+    (void)file; /* squash the unused parameter warning */
     return (ICON_IMAGE)0;
 }
 
@@ -225,12 +239,12 @@ ICON_IMAGE load_icon_file(const char *file) {
 
 /**************************************** client callbacks */
 
-void ui_new_chain(const int section_number) {
-    (void)section_number; /* skip warning about unused parameter */
+void ui_new_chain(const unsigned section_number) {
+    (void)section_number; /* squash the unused parameter warning */
 }
 
-void ui_clients(const int num) {
-    (void)num; /* skip warning about unused parameter */
+void ui_clients(const long num) {
+    (void)num; /* squash the unused parameter warning */
 }
 
 /**************************************** s_log callbacks */
@@ -242,18 +256,13 @@ void ui_new_log(const char *line) {
 /**************************************** ctx callbacks */
 
 int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
-    (void)buf; /* skip warning about unused parameter */
-    (void)size; /* skip warning about unused parameter */
-    (void)rwflag; /* skip warning about unused parameter */
-    (void)userdata; /* skip warning about unused parameter */
-    return 0; /* not implemented */
+    (void)userdata; /* squash the unused parameter warning */
+    return PEM_def_callback(buf, size, rwflag, NULL);
 }
 
-#ifdef HAVE_OSSL_ENGINE_H
-int pin_cb(UI *ui, UI_STRING *uis) {
-    (void)ui; /* skip warning about unused parameter */
-    (void)uis; /* skip warning about unused parameter */
-    return 0; /* not implemented */
+#ifndef OPENSSL_NO_ENGINE
+UI_METHOD *UI_stunnel() {
+    return UI_OpenSSL();
 }
 #endif
 
