@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2016 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -55,7 +55,7 @@
 #define STUNNEL_PLATFORM "Win32"
 #endif
 #define SERVICE_NAME TEXT("stunnel")
-#define SERVICE_DISPLAY_NAME TEXT("Stunnel SSL wrapper")
+#define SERVICE_DISPLAY_NAME TEXT("Stunnel TLS wrapper")
 #endif
 
 /* mingw-Patches-1825044 is missing in Debian Squeeze */
@@ -71,6 +71,7 @@ NOEXPORT void CALLBACK timer_proc(HWND, UINT, UINT_PTR, DWORD);
 NOEXPORT LRESULT CALLBACK window_proc(HWND, UINT, WPARAM, LPARAM);
 NOEXPORT LRESULT CALLBACK about_proc(HWND, UINT, WPARAM, LPARAM);
 NOEXPORT LRESULT CALLBACK pass_proc(HWND, UINT, WPARAM, LPARAM);
+NOEXPORT int pin_cb(UI *, UI_STRING *);
 
 NOEXPORT void save_log(void);
 NOEXPORT void win_log(LPCTSTR);
@@ -110,8 +111,6 @@ static struct LIST {
   TCHAR txt[1]; /* single character for trailing '\0' */
 } *head=NULL, *tail=NULL;
 
-static unsigned number_of_sections=0;
-
 static HINSTANCE ghInst;
 static HWND edit_handle=NULL;
 static HMENU tray_menu_handle=NULL;
@@ -723,7 +722,7 @@ NOEXPORT LRESULT CALLBACK pass_proc(HWND dialog_handle, UINT message,
                 (WPARAM)0 /* line 0 */, (LPARAM)pass_dialog.txt);
             pass_dialog.txt[pass_len]='\0'; /* null-terminate the string */
 
-            /* convert input password to UTF-8 string (as ui_data->pass) */
+            /* convert input passphrase to UTF-8 string (as ui_data->pass) */
             pass_txt=tstr2str(pass_dialog.txt);
             strcpy(ui_data->pass, pass_txt);
             str_free(pass_txt);
@@ -756,7 +755,21 @@ int passwd_cb(char *buf, int size, int rwflag, void *userdata) {
 }
 
 #ifndef OPENSSL_NO_ENGINE
-int pin_cb(UI *ui, UI_STRING *uis) {
+UI_METHOD *UI_stunnel() {
+    static UI_METHOD *ui_method=NULL;
+
+    if(ui_method) /* already initialized */
+        return ui_method;
+    ui_method=UI_create_method("stunnel WIN32 UI");
+    if(!ui_method) {
+        sslerror("UI_create_method");
+        return NULL;
+    }
+    UI_method_set_reader(ui_method, pin_cb);
+    return ui_method;
+}
+
+NOEXPORT int pin_cb(UI *ui, UI_STRING *uis) {
     ui_data=UI_get0_user_data(ui); /* was: ui_data=UI_get_app_data(ui); */
     if(!ui_data) {
         s_log(LOG_ERR, "INTERNAL ERROR: user data data pointer");
@@ -886,7 +899,8 @@ NOEXPORT void daemon_thread(void *arg) {
     SetEvent(main_initialized); /* unlock the GUI thread */
     /* get a valid configuration */
     while(main_configure(cmdline.config_file, NULL)) {
-        cmdline.config_file=NULL; /* don't retry commandline switches */
+        if(cmdline.config_file && *cmdline.config_file=='-')
+            cmdline.config_file=NULL; /* don't retry commandline switches */
         unbind_ports(); /* in case initialization failed after bind_ports() */
         log_flush(LOG_MODE_ERROR); /* otherwise logs are buffered */
         PostMessage(hwnd, WM_INVALID_CONFIG, 0, 0); /* display error */
@@ -967,10 +981,6 @@ NOEXPORT void update_peer_menu(void) {
             DeleteMenu(tray_peer_list, 0, MF_BYPOSITION);
 
     /* initialize data structures */
-    number_of_sections=0;
-    for(section=service_options.next; section; section=section->next)
-        section->section_number=number_of_sections++;
-
     section_number=0;
     for(section=service_options.next; section; section=section->next) {
         servname=str2tstr(section->servname);
